"""CLI interface for qojpca project.
"""
import argparse
#import cupy as np
import numpy as np
from qojpca import base
from pathlib import Path
import os 

def run_qojpca(args):
    P= np.array([])
    Q= np.array([])
    if(len(args.matrices) == 2):
        X,Y = load_matrices(args)
    if(len(args.matrices) == 4):
        X,Y,P,Q = load_matrices(args)
    if args.l_x is None:
        print("INFO: computing "+str(X.shape[1])+" latent variable for the first matrix.")
        l_p = X.shape[1]
    if args.l_y is None:
        print("INFO: computing "+str(Y.shape[1])+" latent variable for the second matrix.")
        l_q = Y.shape[1]
    if(args.l_x is not None):
        l_p = int(args.l_x)
    if(args.l_y is not None):
        l_q = int(args.l_y)
    l = 1.0
    if(args.penalization is not None):
        l = float(args.penalization)
    P_vals,Q_vals,P,Q = base.qojpca(X,Y,l_p,l_q,P,Q,l)
    if(not os.path.isdir(args.output_folder)):
        os.makedirs(args.output_folder)
    np.save(args.output_folder+"/P.npy",P)
    np.save(args.output_folder+"/P_vals.npy",P_vals)
    np.save(args.output_folder+"/Q.npy",Q)
    np.save(args.output_folder+"/Q_vals.npy",Q_vals)

    
def run_jpca(args):
    X,Y =load_matrices(args)
    if args.l_x is None:
        print("INFO: computing "+str(X.shape[1])+" latent variable for the first matrix.")
        l_x = X.shape[1]
    if args.l_y is None:
        print("INFO: computing "+str(Y.shape[1])+" latent variable for the second matrix.")
        l_y = Y.shape[1]
    if(args.l_x is not None):
        l_x = int(args.l_x)
    if(args.l_y is not None):
        l_y = int(args.l_y)
        
    P_vals,Q_vals,P,Q = base.jpca(X,Y,l_x,l_y)
    if(not os.path.isdir(args.output_folder)):
        os.makedirs(args.output_folder)
    np.save(args.output_folder+"/P.npy",P)
    np.save(args.output_folder+"/P_vals.npy",P_vals)
    np.save(args.output_folder+"/Q.npy",Q)
    np.save(args.output_folder+"/Q_vals.npy",Q_vals)
    
def load_matrices(args):
    matrices = args.matrices
    if(len(matrices) == 2):
        X = np.load(matrices[0])
        Y = np.load(matrices[1])
        print("INFO: matrices successfully loaded")
        return X,Y
    if(len(matrices) == 4):
        X = np.load(matrices[0])
        Y = np.load(matrices[1])
        P = np.load(matrices[2])
        Q = np.load(matrices[3])
        print("INFO: matrices successfully loaded")
        return X,Y,P,Q
    print("ERROR: wrong number of matrices provided")
    return 0

    
def main():  # pragma: no cover
    """
    The main function executes on commands:
    `python -m qojpca` and `$ qojpca `.
    """
    parser = argparse.ArgumentParser(description='QOJPCA command line interface.')
    parser.add_argument('matrices', metavar='M', type=str, nargs="+",
                    help='matrices to be processed')
    parser.add_argument('--qojpca', action='store_true',
                    help='performs QOJPCA on given matrices')
    parser.add_argument('--jpca', action='store_true',
                    help='performs JPCA on given matrices')
    parser.add_argument('--l_x', 
                    help='number of latent variables to compute for the first matrix, default is 50')
    parser.add_argument('--l_y',
                    help='number of latent variables to compute for the second matrix, default is 50')
    parser.add_argument('--output_folder',required=True,
                    help='output directory')
    parser.add_argument('--penalization',
                    help='penalization value, that is lambda in the paper')              
    args = parser.parse_args()
    
    
    if(args.qojpca):
        run_qojpca(args)

    if(args.jpca):
        run_jpca(args)

    



"""
qojpca base module.

"""
from scipy.sparse.linalg import eigsh
from scipy.sparse.linalg import LinearOperator
#import cupy as np
import numpy as np

l= 1.0
k = 1.0
X=np.array([])
Y=np.array([])
P=np.array([])
Q=np.array([])

def logError(message):
    print("@@@@@")
    print("ERROR: "+message)
    print("@@@@@")
def logInfo(message):
    print("INFO: "+message)
    

def mv1_it1(v):
    return k*X.dot(X.transpose().dot(v)) 

def mv2_it1(v):
    return Y.dot(Y.transpose().dot(v))

def stepP(v):
    return k*X.dot(X.transpose().dot(v)) - l * Q.transpose().dot(Q.dot(v))

def stepQ(v):
    return Y.dot(Y.transpose().dot(v)) - l * P.transpose().dot(P.dot(v))


def qojpca(X_,Y_,l_x, l_y,P_=np.array([]),Q_=np.array([]), l_ = 1.0 , nb_iterations=10, relative_convergence_criteria= 0.05):
    global P,Q,k,l,X,Y
    i=0
    
    if(X_.shape[0]==Y_.shape[0]):
        n = X_.shape[0]
        X=X_
        Y=Y_
        k= X_.shape[1]/X_.shape[1]
        l=l_
        P=P_
        Q=Q_
        if(P.shape[0]==0 or Q.shape[0]==0):
            P_vals,Q_vals,P,Q = jpca(X_,Y_,l_x,l_y)

            logInfo("multiplying the penalization term by the largest eigenvalue of kXX^T, i.e.:" +str(P_vals[0]))
            
            l = l*P_vals[0]
        
        samplingFactor = 40  
        samples  = n/40
        colsIds = np.random.choice(n, int(samples),replace=False)
        
        PQT = P.dot(Q.transpose())
        orthoScore= np.linalg.norm(PQT)
        Xscore = np.linalg.norm(X[colsIds,:] - P[:,colsIds].transpose().dot(P[:,colsIds]).dot(X[colsIds,:]))
        Yscore = np.linalg.norm(Y[colsIds,:] - Q[:,colsIds].transpose().dot(Q[:,colsIds]).dot(Y[colsIds,:]))        
        totalEnergy = k*Xscore*samplingFactor + Yscore*samplingFactor + np.sqrt(l)*orthoScore
        
        
        for i in range(1,nb_iterations): 
            G2 = LinearOperator((n,n), matvec=stepQ)
            Q_vals,Q = eigsh(G2, k=l_y,which ='LA')
            Q = Q.transpose()
    
            G1 = LinearOperator((n,n), matvec=stepP)
            P_vals,P = eigsh(G1, k=l_x, which ='LA')
            P = P.transpose()
            
            PQT = P.dot(Q.transpose())
            orthoScore = np.sqrt(l)*np.linalg.norm(PQT)
            
            Xscore = np.linalg.norm(X[colsIds,:] - P[:,colsIds].transpose().dot(P[:,colsIds]).dot(X[colsIds,:]))
            Yscore = np.linalg.norm(Y[colsIds,:] - Q[:,colsIds].transpose().dot(Q[:,colsIds]).dot(Y[colsIds,:]))        
            newTotalEnergy = k*Xscore*samplingFactor + Yscore*samplingFactor + orthoScore
            logInfo("X score ="+ str(k*Xscore*samplingFactor)+" | Y score ="+ str(Yscore*samplingFactor)+" | orthogonallity score ="+ str(orthoScore)+" |  total energy ="+ str(newTotalEnergy))
            

            if(abs(totalEnergy - newTotalEnergy)/ newTotalEnergy < relative_convergence_criteria ):
                break
            totalEnergy = newTotalEnergy
        
        return P,Q,P_vals,Q_vals
    else:
        logError("dimension mismatch between X and Y")
        return np.array(),np.array(),np.array(),np.array()
        
    


def jpca(X_,Y_,l_p,l_q):
    global X,Y,k
    
    if(X_.shape[0]==Y_.shape[0]):
        n = X_.shape[0]
        X=X_
        Y=Y_
        k=  Y_.shape[1]/X_.shape[1]
        logInfo("Computing k normalization automatically by adjusting the number of samples, k=: "+ str(k))
    else:
        print("ERROR: dimension mismatch between X and Y")
        return 0
    G1_it1 = LinearOperator((n,n), matvec=mv1_it1)
    P_vals,P = eigsh(G1_it1, k=l_p,which ='LA')
    G2_it1 = LinearOperator((n,n), matvec=mv2_it1)
    Q_vals,Q = eigsh(G2_it1, k=l_q,which ='LA')
    P = P.transpose()
    Q = Q.transpose()
    return P_vals,Q_vals,P,Q


"""Python setup.py for qojpca package"""
import io
import os
from setuptools import find_packages, setup


def read(*paths, **kwargs):
    """Read the contents of a text file safely.
    >>> read("qojpca", "VERSION")
    '0.1.1'
    >>> read("README.md")
    ...
    """

    content = ""
    with io.open(
        os.path.join(os.path.dirname(__file__), *paths),
        encoding=kwargs.get("encoding", "utf8"),
    ) as open_file:
        content = open_file.read().strip()
    return content


def read_requirements(path):
    return [
        line.strip()
        for line in read(path).split("\n")
        if not line.startswith(('"', "#", "-", "git+"))
    ]


setup(
    name="qojpca",
    version=read("qojpca", "VERSION"),
    description="qojpca package",
    url="https://github.com/florentjousse/qojpca/",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="florentjousse",
    packages=find_packages(exclude=["tests", ".github"]),
    install_requires=['numpy','scikit-learn'
    ],
    entry_points={
        "console_scripts": ["qojpca = qojpca.__main__:main"]
    },
    extras_require={"test": read_requirements("requirements-test.txt"), "gpu": 'cupy-cuda11x'},
)
